#
# $Id: events.py,v 1.19 2006/05/19 23:50:35 mwalfish Exp $
#
# Simple event loop. Basically copying the relevant pieces of Dave
# Mazieres's async/core.C. Obviously there are many ways to build an event
# loop; we chose Dave's because, well, it was there and it works in C++.
#
#
# Copyright (C) 2005-2006 Michael Walfish
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import signal
import select
import errno
import time
import dqm.err as dqm_err
import dqm.rpc.fifoq as fifoq

READ  = 0
WRITE = 1

# globals
fd_sets     = { READ:{} , WRITE:{} }
fifo_q      = fifoq.fifoq()
short_lst   = []
selwait     = 0.0
stop        = False
loop_called = False

# We have to return a handle to users so that (when,callback) pairs
# that have multiplicities > 1 are unique. The equality test is then 
# on _which_ handle not on the pair (when,callback).
class timercb_handle:
    
    def __init__(self, when, callback, arg=None):
	self.when = when
	self.callback = callback
	self.arg    = arg
	self.active = True
   
# Register to have 'callback' called when file descriptor
# 'fd' is ready for 'type', which is either READ or WRITE.
# Callback is a function that takes no arguments. 
def reg_fd_cb(fd, type, callback):

    fd_sets[type][fd] = callback

# It is an error to call unreg on (fd,type) if you have 
# not previously called reg on (fd,type).
def unreg_fd(fd, type):

    del fd_sets[type][fd]


# 'delta' is a number of seconds or fractions thereof. it is a floating
# point number. we let callback take an argument instead of using lambdas
# because lambdas are expensive in python, so leave the choice up to
# our callers.
def reg_fifo_timer_delay(delta, callback, arg=None):
   
    when = delta + time.time()
    handle = reg_fifo_timer_at(when, callback, arg)
    return handle

#  'when' is a floating point number representing the number
# of seconds since Jan 1, 1970.
def reg_fifo_timer_at(when, callback, arg=None):

    handle = timercb_handle(when, callback, arg)
    fifo_q.append(handle)
    return handle

def reg_long_timer_delay(delta, callback, arg=None):

    when = delta + time.time()
    handle = reg_long_timer_at(when, callback, arg)
    return handle

def reg_long_timer_at(when, callback, arg=None):

    handle = timercb_handle(when, callback, arg)
    for i in range(len(short_lst)):
	if when < short_lst[i].when:
	    short_lst.insert(i, handle)
	    break
    else:
	short_lst.append(handle)

# It is an error to deregister a handle that wasn't returned 
# from one of the reg_timer_* functions
def unreg_timer(handle):
    handle.active = False

cntr = 0
def check_timers():

#    global cntr
#    cntr = cntr + 1
#    if cntr % 2000 == 0: 
#	print "fifoq: %d. short list: %d" % (len(fifo_q), len(short_lst))

    now = time.time()
    handle = fifo_q.front()
    while True:

	while handle and handle.when <= now:

	    fifo_q.pop()
	    if handle.active:
		if handle.arg: handle.callback(handle.arg)
		else:          handle.callback()

	    handle = fifo_q.front()

	# update our current conception of time and go through
	# the inner loop again.
	now = time.time()
	if handle is None or now < handle.when:
	    break

    # now check the short list. don't update time. it's okay if the
    # items on this list are late.
    while len(short_lst) > 0 and short_lst[0].when <= now:

	handle = short_lst[0]
	short_lst.pop(0)

	if handle.active:
	    if handle.arg: handle.callback(handle.arg)
	    else:	   handle.callback()


    global selwait
    first = None

    # Pick the smallest from the front of the two lists.
    if len(fifo_q): 
	first = fifo_q.front().when
    
    if len(short_lst) > 0 and \
          ((first is None) or (short_lst[0].when < first)):
	first = short_lst[0].when

    if first is None:
	selwait = 60.0
    else:
	now = time.time()
	if first < now: selwait = 0
	else:           selwait = first - now

def check_fdcbs():

    rd_set = fd_sets[READ]
    wr_set = fd_sets[WRITE]
    rd_fds = rd_set.keys()
    wr_fds = wr_set.keys()

    try:

	readable,writable,d = select.select(rd_fds, wr_fds, [], selwait)

	for fd in readable:
	    rd_set[fd]()

	for fd in writable:
	    wr_set[fd]()

    except select.error, e:

	if e[0] != errno.EINTR: 
	    print 'selwait: %s' % str(selwait)
	    raise dqm_err.Error('errno %s: %s' % (errno.errorcode[e[0]], e[1]))


def stop_looping():

    global stop, loop_called

    loop_called = False
    stop = True

def checks():

    check_timers()
    # have to check 'stop' again b/c if the caller asked us to stop, it 
    # may be because there are no more events the caller cares about, in
    # which case the select call in check_fdcbs() would wait forever.
    if stop is False: 
	check_fdcbs()
    # XXX: signals?

def loop(override_cancel_handler=True):

    global loop_called, stop

    if loop_called: raise dqm_err.Error("loop() called twice")

    loop_called = True
    stop = False

    if override_cancel_handler:
	signal.signal(signal.SIGINT, cancel_handler)
    while stop is False:
	checks()

def cancel_handler(signal, frame):
    stop_looping()
    # keep pychecker happy
    a = str(signal) + str(frame)
    a = a[8:8]
