#
# $Id: xdr_rpc.py,v 1.9 2006/05/19 23:50:35 mwalfish Exp $
#
# Routines from Sameer and Emil to handle XDR RPC request and reply
# messages. 
# 
#
#
# Copyright (C) 2005-2006 Michael Walfish
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import dqm.svc.RPCProto as RPCProto
from xdrlib import Packer, Unpacker

class UnpackException(Exception):
    """An exception that may indicate which xid"""
    def __init__(self,st,xid=None):
	Exception.__init__(self)
	self.st  = st
	self.xid = xid
    def __str__(self):
	return self.st
	

class ReplyException(Exception):
    """An exception that carries a reply packet"""
    def __init__(self, message, reply):
        Exception.__init__(self)
        self.message = message
        self.reply = reply
    def __str__(self):
        return self.message

def check(expected, actual, name, replyf=None, xid=None):
    """If expected is not None, checks whether expected equals actual,
    and if not, raises an exception (see code for themessage).
    If replyf is None, the exception is an UnpackException.
    Otherwise, reply must be a function that takes no arguments
    and returns an RPC reply (a string), and the exception is a
    ReplyException containing the output of the function.
    """
    if expected is not None and expected != actual:
        if replyf:
            raise ReplyException("Expected %s %s, but got %s"
                                  % (name, expected, actual), replyf())
        else:
            raise UnpackException("Expected %s %s, but got %s"
                                  % (name, expected, actual), xid)

def pack_reply(xid, *args):
    """Packs an RPC reply from a variable-length arg list (args):
    MSG_ACCEPTED, verf, (SUCCESS | PROG_MISMATCH, low, hi | PROG_UNAVAIL
                         | PROC_UNAVAIL | GARBAGE_ARGS | SYSTEM_ERR)
    MSG_DENIED, (RPC_MISMATCH, hi, low | AUTH_ERROR, auth_stat)

    verf is an auth of the form (flavor, value)
    Returns an xdrlib.Packer that the caller can use to add data,
    such as the results of a RPCProto.SUCCESSful call.
    """
    arg = list(args) # need a mutable list for pop()
    msg = RPCProto.rpc_msg()
    msg.xid = xid
    msg.body = RPCProto.body_t()
    msg.body.mtype = RPCProto.REPLY
    msg.body.rbody = reply = RPCProto.reply_body()
    reply.stat = reply_stat = arg.pop(0)
    if reply_stat == RPCProto.MSG_ACCEPTED:
        reply.areply = RPCProto.accepted_reply()
        reply.areply.reply_data = RPCProto.reply_data_t()
        reply.areply.reply_data.stat = accept_stat = arg.pop(0)
        if accept_stat == RPCProto.PROG_MISMATCH:
            reply.areply.reply_data.mismatch_info = RPCProto.mismatch_info_t()
            reply.areply.reply_data.mismatch_info.low  = arg.pop(0)
            reply.areply.reply_data.mismatch_info.high = arg.pop(0)
        elif (accept_stat == RPCProto.SUCCESS):
            reply.areply.reply_data.results = '' # FIXME?
        elif (accept_stat == RPCProto.PROG_UNAVAIL or
              accept_stat == RPCProto.PROC_UNAVAIL or
              accept_stat == RPCProto.GARBAGE_ARGS or
              accept_stat == RPCProto.SYSTEM_ERR):
            pass
        else:
            raise ValueError("unknown accept_stat: %u" % accept_stat)
    elif reply_stat == RPCProto.MSG_DENIED:
        reply.rreply = RPCProto.rejected_reply()
        reply.rreply.stat = reject_stat = arg.pop(0)
        if reject_stat == RPCProto.RPC_MISMATCH:
            reply.rreply.mismatch_info.low = RPCProto.mismatch_info_t()
            reply.rreply.mismatch_info.low = arg.pop(0)
            reply.rreply.mismatch_info.high = arg.pop(0)
        elif reject_stat == RPCProto.AUTH_ERROR:
            reply.rreply.astat = arg.pop(0)
        else:
            raise ValueError("unknown reject_stat: %u" % reject_stat)
    else:
        raise ValueError("unknown reply_stat: %u" % reply_stat)
    p = Packer()
    RPCProto.pack_rpc_msg(p, msg)
    return p

def unpack_reply(response, myxid=None, myreply_stat=RPCProto.MSG_ACCEPTED,
                 myaccept_stat=RPCProto.SUCCESS,
                 myreject_stat=None, myauth_stat=None):
    """Unpacks an RPC reply and returns a variable-length arg list
    of the same form as the argument to pack_reply, but for RPCProto.SUCCESS also
    returns an xdrlib.Unpacker as the final element of the list
    that the caller can use to unpack the results of the call.

    If values are given for any myYYY arguments, checks that those
    values match the unpacked YYY values.  Default myYYY values assume
    success with no authentication.
    
    Raises UnpackException on any errors or mismatches.
    """
    u = Unpacker(response)
    msg = RPCProto.unpack_rpc_msg(u)
    check(myxid, msg.xid, "xid")
    if msg.body.mtype == RPCProto.CALL:
        raise UnpackException("Expected reply, but got call", xid=msg.xid)
    reply = msg.body.rbody
    check(myreply_stat, reply.stat, "reply_stat", xid=msg.xid)
    retval = [msg.xid, reply.stat]
    if reply.stat == RPCProto.MSG_ACCEPTED:
#        check(myverf, reply.areply.verf, "verf", xid=msg.xid)
#        retval.append(reply.areply.verf)
        accept_stat = reply.areply.reply_data.stat
        check(myaccept_stat, accept_stat, "accept_stat", xid=msg.xid)
        retval.append(accept_stat)
        if accept_stat == RPCProto.SUCCESS:
            retval.append(u)
        elif accept_stat == RPCProto.PROG_MISMATCH:
            retval.append(reply.areply.reply_data.mismatch_info.low)
            retval.append(reply.areply.reply_data.mismatch_info.high)
        elif (accept_stat == RPCProto.PROG_UNAVAIL or
              accept_stat == RPCProto.PROC_UNAVAIL or
              accept_stat == RPCProto.GARBAGE_ARGS or
              accept_stat == RPCProto.SYSTEM_ERR):
            pass
        else:
            raise UnpackException("unknown accept_stat: %u" % accept_stat,
				  xid=msg.xid)
    elif reply.stat == RPCProto.MSG_DENIED:
        reject_stat = reply.rreply.stat
        check(myreject_stat, reject_stat, "reject_stat", xid=msg.xid)
        retval.append(reject_stat)
        if reject_stat == RPCProto.RPC_MISMATCH:
            retval.append(reply.rreply.mismatch_info.low)
            retval.append(reply.rreply.mismatch_info.high)
        elif reject_stat == RPCProto.AUTH_ERROR:
            check(myauth_stat, reply.rreply.astat, "auth_stat", xid=msg.xid)
            retval.append(reply.rreply.astat)
        else:
            raise UnpackException("unknown reject_stat: %u" % reject_stat,
				  xid=msg.xid)
    else:
        raise UnpackException("unknown reply_stat: %u" % reply.stat,
			      xid=msg.xid)
    return retval
    
def pack_call(xid, prog, vers, proc):
    """Packs an RPC call message; returns an xdrlib.Packer that
    the caller can use to add more data, e.g., the call arguments.
    """
    msg = RPCProto.rpc_msg()
    msg.xid = xid
    msg.body = RPCProto.body_t()
    msg.body.mtype = RPCProto.CALL
    msg.body.cbody = RPCProto.call_body()
    msg.body.cbody.rpcvers = RPCProto.RPC_VERSION
    msg.body.cbody.prog = prog
    msg.body.cbody.vers = vers
    msg.body.cbody.proc = proc
    p = Packer()
    RPCProto.pack_rpc_msg(p, msg)
    return p

def unpack_call(request, myprog=None, myvers=None, myprocs=None):
    """Unpacks an RPC call message from request.

    Returns (xid, prog, vers, proc, cred, verf, u) if okay,
    where u is an xdrlib.Unpacker.
    otherwise raises either UnpackException or ReplyException.
    If myYYY is not None, checks that YYY == myYYY.
    Assumes AUTH_NONE for cred and verf; override with mycred and myverf.
    """
    if len(request) < 24:
        raise UnpackException("Packet too short (%d bytes)" % len(request))
    u = Unpacker(request)
    msg = RPCProto.unpack_rpc_msg(u)
    if msg.body.mtype == RPCProto.REPLY:
        raise UnpackException("Expected call, but got reply")
    call = msg.body.cbody
    check(RPCProto.RPC_VERSION, call.rpcvers, "RPC version",
          lambda: pack_reply(msg.xid,
                             RPCProto.MSG_DENIED,
                             RPCProto.RPC_MISMATCH,
                             RPCProto.RPC_VERSION,
                             RPCProto.RPC_VERSION).get_buffer())
    check(myprog, call.prog, "program",
          lambda: pack_reply(msg.xid,
                             RPCProto.MSG_ACCEPTED,
                             RPCProto.PROG_UNAVAIL).get_buffer())
    check(myvers, call.vers, "version",
          lambda: pack_reply(msg.xid,
                             RPCProto.MSG_ACCEPTED,
                             RPCProto.PROG_MISMATCH,
                             myvers,
                             myvers).get_buffer())

    # MW hacking in b/c the test belongs here even though the 'check()'
    # function isn't built to handle tests of this sort
    if myprocs is not None and call.proc not in myprocs:
	# hack this somewhat-hoopty check interface which only
	# accepts tests of the form (a == b)
	check(call.proc + 100, call.proc, "proc",
	     lambda: pack_reply(msg.xid,
				RPCProto.MSG_ACCEPTED,
				RPCProto.PROC_UNAVAIL).get_buffer())

#    check(mycred, call.cred, "cred",
#          lambda: pack_reply(msg.xid,
#                             RPCProto.MSG_DENIED,
#                             RPCProto.AUTH_ERROR,
#                             RPCProto.AUTH_BADCRED).get_buffer())
#    check(myverf, call.verf, "verf",
#          lambda: pack_reply(msg.xid,
#                             RPCProto.MSG_DENIED,
#                             RPCProto.AUTH_ERROR,
#                             RPCProto.AUTH_BADVERF).get_buffer())
#    return (msg.xid, call.prog, call.vers,
#            call.proc, call.cred, call.verf, u)
    return (msg.xid, call.prog, call.vers, call.proc, u)
