#
# $Id: q.py,v 1.7 2006/05/19 23:50:35 mwalfish Exp $ 
#  
#  On-disk to-do list stored as (key,value) pairs. The keys and values are
#  python strings. Underneath, this object uses the Berkeley DB.
#
#
# Copyright (C) 2005-2006 Michael Walfish
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
import os
import os.path
import time
import cPickle
import bsddb3.db as bdb
import dqm.logger

class Error(Exception):
    def __init__(self, st=None):
	Exception.__init__(self)
	self.st = st
    def __str__(self):
	if self.st is None: return ''
	else:               return self.st

def output_db_err(errval, logger):
    rc,msg = errval
    logmsg = "error: " + msg + ". code: " + str(rc)
    logger.log(logmsg)

"""On-disk queue that exports a few simple methods."""
class Queue:
  
    EXTENT_SIZE = 64*1024
    ENV_NEW_FL  = bdb.DB_INIT_LOCK  | bdb.DB_INIT_TXN  | bdb.DB_INIT_LOG |\
	          bdb.DB_INIT_MPOOL | bdb.DB_CREATE    | bdb.DB_RECOVER    
    ENV_JOIN_FL = bdb.DB_JOINENV
    DB_FLAGS    = bdb.DB_CREATE

    def __init__(self, \
		 dir, \
		 name, \
		 custodian=False, \
		 logger=None, \
		 rec_len=2048, \
                 check_interval=2,
		 mode=0770):

	"""Construct an on-disk queue. Arguments:
	    dir: directory in which the queue will be created or already
	         lives
	    filename: name of the queue within the directory
	    custodian: if the database already exists, then we will
	               only run recovery when 'custodian' is True.
		       if the databasedir doesn't exist yet, the value of
		       custodian is irrelevant. set custodian to be true if
		       you are a parent process that is responsible for
		       recovering the database. Note that this means that
		       *if* the database already exists and *if* a custodian has
		       not run on a damaged database, *then* the database
		       may not open properly or may hang. It also means
		       that if the database dir *does* exist but the
		       database itself hasn't been created, then this class
		       will raise an exception.
	    logger: an object that exports a 'log' method
	    rec_len: length of the record in the database. It is an error
		     to enqueue() something that becomes larger than
		     rec_len. Note that before enqueuing, the (key,value)
		     pairs are pickled and will thus get a little bigger.
	    check_interval: in seconds, the sleep interval between polls of
		     the database by the 'dequeue' method.
	"""

	self.closed = False
	self.check_interval = check_interval
	self.db = None
	self.db_env = None

	if logger is None: logger = dqm.logger.stderr_logger()
	self.logger = logger

	exist = os.path.isdir(dir)

	if custodian:
	    flags = self.ENV_NEW_FL
	    if not exist: os.makedirs(dir, mode)
	else:
	    flags = self.ENV_JOIN_FL
	    if not exist: logger.log("DB doesn't exist"); raise Error()

	# zero out all of the 'execute' bits
	mode = mode & ~0111
	self.db_env = bdb.DBEnv()
	try:
	    self.db_env.open(dir, flags, mode)
	    self.db = bdb.DB(self.db_env)

	    self.db.set_re_len(rec_len)
	    self.db.set_q_extentsize(self.EXTENT_SIZE)

	    # NOTE: later transactions will fail if we don't surround the DB
	    # open with a transaction. 
	    txn = self.db_env.txn_begin()
	    self.db.open(name, dbtype=bdb.DB_QUEUE, flags=self.DB_FLAGS, 
			 txn=txn, mode=mode)
	    txn.commit()

	except bdb.DBRunRecoveryError, errval:
	    output_db_err(errval, self.logger)
	    self.logger.log("standard catastrophe.")
	    raise Error(errval[1])

        except bdb.DBError, errval:
	    output_db_err(errval, self.logger)
	    raise Error(errval[1])

    def __del__(self):
	if not self.closed: self.close()

    def close(self):
	if self.db is not None:     self.db.close()
	if self.db_env is not None: self.db_env.close()
        del self.db
	del self.db_env
	self.closed = True


    def enqueue(self, key, value, block=True, max_tries=5):
	"""key and value should be Python strings. If block=True, then enqueue 
	will always return True or else throw
	an "Error" exception (if something went badly wrong). If this function
	throws an exception, the caller should exit because the Berkeley DB
	needs to be restarted. If block=False, then enqueue will either
	return 'True' or 'False'. 'True' means that the key,value pair was
	enqueued. 'False' means the key,value pair wasn't enqueued.
	'max_tries' only counts if the thing isn't blocking"""

        done = False
	rc   = False

	record = self.marshall(key,value)
	tries = 0

        while not done:

	    txn = self.db_env.txn_begin()
	    try:
		self.db.append(record, txn)	# ignore return value
		txn.commit()
		done = True
		rc   = True

	    except bdb.DBLockDeadlockError, errval:
		txn.abort()
		tries = tries + 1
		if not block and tries >= max_tries: done = True

	    except bdb.DBRunRecoveryError, errval:
		output_db_err(errval, self.logger)
		self.logger.log("standard catastrophe.")
		raise Error(errval[1])

	    except bdb.DBError, errval:
		output_db_err(errval, self.logger)
		raise Error(errval[1])

        return rc


    def dequeue(self, block=True):
	"""If block=True, then dequeue will always return a key,value
	pair or else throw an 'Error' exception (if something went badly
	wrong). See comment above about the exception. If block=False, then
	dequeue will either return a (key,value) tuple or None. 'None' means
	the queue was empty. 'key' and 'value' are python strings."""

	done   = False
	record = None

	while not done:

	    txn = self.db_env.txn_begin()
	    try: 
		record = self.db.consume(txn=txn, flags=bdb.DB_RMW)
		txn.commit()
		if record is not None or block is False: done = True

	    except bdb.DBLockDeadlockError, errval:
		txn.abort()
		if not block: done = True

	    except bdb.DBRunRecoveryError, errval:
		output_db_err(errval, self.logger)
		self.logger.log("standard catastrophe.")
		raise Error(errval[1])

	    except bdb.DBError, errval:
		output_db_err(errval, self.logger)
		raise Error(errval[1])

	    if record is None: time.sleep(self.check_interval)
	    

	if record is not None: 
	    record = record[1].strip() # berkeley DB pads with ' '
	    key,val = self.unmarshall(record)
	    record = (key,val)

	return record


    def marshall(self, key, value):
	return cPickle.dumps((key,value)) 

    def unmarshall(self, record):
	key, value = cPickle.loads(record)
	return key, value
