/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#include <assert.h>
#include <unistd.h>
#include "arpc.h"
#include "dqm_storer.h"
#include "dqm_node.h"
#include "config.h"
#include "nodetable_geom.h"
#include "nodetable_notree.h"
#include "storer.h"
#include "portal.h"
#include "util.h"
#include "log.h"
#include "lru_cache.h"
#include "ondisk_storage.h"

void usage() 
{
    warnx << "Usage: " << progname 
          << " [-l logfile]"
	  << " [-p portal_port]"
	  << " [-s storer_port]\n"
	  << " [-i internal_port]"
	  << " [-d key_val_storage]"
	  << " [-t use_tree]\n"
	  << " [-k num_rings]"
	  << " [-r tree_degree]"
	  << " [-f repl_factor]\n"
	  << " [-c cache_size]"
	  << " [-n index_entries]"
	  << " [-h help_msg]"
	  << " config file\n";
    warnx << "  -l: if no logfile given, stderr is used\n";
    warnx << "  -p: portal's listening port. "
             "overrides the value in the config file\n";
    warnx << "  -s: storer's listening port. "
             "overrides the value in the config file\n";
    warnx << "  -i: internal port (source UDP port for traffic between "
             "nodes).\n      overrides the value in the config file.\n";
    warnx << "  -d: if given, overrides the config file\n";
    warnx << "  -t: use a tree or the non-tree configuration.\n"
             "      overrides the value in the config file\n";
    warnx << "  -k: number of rings. "
	     "if given, overrides the value in the config file\n";
    warnx << "  -r: degree of the tree. "
	     "if given, overrides the value in the config file\n";
    warnx << "  -f: repl_factor for non-tree configuration (r).\n"
             "      overrides the value in the config file\n";
    warnx << "  -c: size of LRU cache holding <k,v> pairs\n"
             "      overrides the value in the config file\n";
    warnx << "  -n: number of entries in the in-memory map from keys to "
                    "disk blocks\n"
             "      overrides the value in the config file\n";
    warnx << "  -h: this message\n";
    exit(1);
}

void
init_nodetable(nodetable* nt, const approved_list* storers, 
	       const str& our_identity, ptr<base_logger> logger) 
{
    approved_list::slot* s;

    for (s = storers->first(); s; s = storers->next(s)) {

	ip_port my_ip_port;
	my_ip_port.ip   = s->value.ip_addr;
	my_ip_port.port = htons(s->value.storer_port);

	nt->add_node (my_ip_port, s->key);
	nt->set_state(my_ip_port, UP);

	if (s->key == our_identity) 
	    nt->set_local_node(my_ip_port);
    }
    nt->output_params(logger);
}

node_id
get_local_id(config_params* params)
{
    node_id local_ip_port;
    const approved_list* app_list = params->get_storers();
    const storer_info* info  = (*app_list)[params->identity()];

    local_ip_port.ip   = info->ip_addr;
    local_ip_port.port = htons(info->storer_port);

    return local_ip_port;
}


class worker {
    public:
	worker(int argc, char** argv);
	~worker();

    private:
	ptr<base_logger> logger;
	ptr<portal> por;
        ptr<storer> sto;
	bool can_exit;
};

worker::worker(int argc, char** argv)
: logger(NULL), por(NULL), sto(NULL), can_exit(false)
{
    int opt;
    ptr<nodetable> nodetbl = NULL;
    ptr<aclnt> my_clnt = NULL;
    ptr<lru_cache> cache = NULL;
    ptr<ondisk_storage> ondisk = NULL;
   
    int32_t read_size = -1, write_size = -1;
    u_int16_t portal_port = 0, 
	      storer_port = 0, 
	      internal_port = 0;
    str storage_name("");
    int num_rings = -1,
	tree_degree = -1;
    int use_tree = -1;
    int r = -1;
    int cache_size = -1;
    uint32_t nindex = 0;

    logger = NULL;

    opterr = 0;
    while ( (opt = getopt(argc, argv, "l:p:s:i:d:k:t:r:f:c:n:h")) != -1) {

	switch (opt) {
	
	    case 'l':
		logger = New refcounted<file_logger>(optarg);	
		break;

	    case 'p':
		portal_port = atoi(optarg);
		break;
   
	    case 's':
		storer_port = atoi(optarg);
		break;

	    case 'i':
		internal_port = atoi(optarg);
		break;

	    case 'd':
		storage_name = optarg;
		break;

	    case 'k':
		num_rings = atoi(optarg);
		break;
	    
	    case 't':
		use_tree = atoi(optarg);
		break;

	    case 'r':
		tree_degree = atoi(optarg);
		break;

	    case 'f':
		r = atoi(optarg);
		break;

	    case 'c':
		cache_size = atoi(optarg);
		break;

	    case 'n':
		nindex = atoi(optarg);
		break;

	    case 'h':
		usage();
		break;

	    default:
		usage();
	}

    }

    if (optind != argc - 1)
	usage();

    config_params conf_params(argv[optind]);
    node_id local_id;

    /* XXX: other signals for debugging. e.g.:
         sigcb(SIGUSR1, wrap (&stats));
	 sigcb(SIGUSR2, wrap (&stop));
	 sigcb(SIGHUP, wrap (&start_logs)); */

    if (logger == NULL)
	logger = New refcounted<stderr_logger>();

    if (portal_port == 0)
	portal_port = conf_params.portal_port();

    if (storer_port == 0)
	storer_port = conf_params.storer_port();

    if (internal_port == 0)
	internal_port = conf_params.internal_port();

    if (storage_name == "")
	storage_name = conf_params.key_val_storage();

    if (num_rings == -1)
	num_rings = conf_params.num_rings();

    if (tree_degree == -1)
	tree_degree = conf_params.tree_degree();

    if (use_tree == -1) 
	use_tree = conf_params.use_tree();

    if (r < 0)
	r = conf_params.r();

    if (cache_size < 0)
	cache_size = conf_params.cache_size();

    if (nindex == 0)
	nindex = conf_params.nindex_entries();

    warnx("bunker-assigned name: %s. port: %d.\n", 
	  conf_params.identity().cstr(), portal_port);

    my_clnt = init_udp_client(internal_port, dqm_storer_program_1);
 
    if (use_tree) {
	if (num_rings < 0 || tree_degree < 0) 
	    fatal("tree requested, but num_rings or tree_degree not "
	          "supplied. (num_rings,tree_degree) = (%d,%d)\n",
		  num_rings, tree_degree);
	nodetbl = New refcounted<nodetable_geom> (tree_degree, num_rings);
	warnx("using tree configuration. nrings: %d. degree: %d\n", 
	      num_rings, tree_degree);
    } else {
	if (r < 0) fatal("r not supplied.\n");
	nodetbl = New refcounted<nodetable_notree>(r);
	warnx("using non-tree configuration. ");
	warnx("r: %d\n", r);
    }

    warnx("LRU cache size: %d. specified index size: %d\n", 
	  cache_size, nindex);

    init_nodetable(nodetbl, conf_params.get_storers(), conf_params.identity(),
                   logger);

    local_id   = get_local_id(&conf_params);

    read_size  = conf_params.read_size();
    write_size = conf_params.write_size();

    cache = New refcounted<lru_cache> (cache_size);
    if (nindex == 0)
	ondisk = New refcounted<ondisk_storage> (storage_name, logger,
						 read_size, write_size);
    else
	ondisk = New refcounted<ondisk_storage> (storage_name, logger,
						 read_size, write_size,
						 nindex);
	
    sto = New refcounted<storer> (storer_port, nodetbl, ondisk, cache, logger);
    por = New refcounted<portal> (conf_params.get_storers(), 
                                  portal_port,
				  my_clnt, nodetbl, cache, sto, local_id, 
				  logger, &can_exit);

    logger->log("beginning....");
    warnx << "beginning\n";
}

worker::~worker()
{
    por->stop();
    sto->stop();

    // release the reference now. don't wait until the ptr's destructor
    // is called. this is because the portal won't exit until all
    // references to it have gone away.
    por = NULL;
    sto = NULL;

    warnx << "waiting for portal to exit . . .\n";
    while (!can_exit)  {
	acheck();
    }
    warnx << "portal exited\n";
}


// The purpose of the worker object is to get constructor/destructor
// semantics for the objects that we create at first. main never exits,
// so destructors for objects allocated in main will never be called.
worker* main_worker = NULL;

void halt()
{
    warnx << "beginning shutdown ....\n";
    delete main_worker;
    exit(0);
}

int main(int argc, char** argv)
{
    srand(time(NULL));
    setprogname(argv[0]);
    //sfs_set_clock(SFS_CLOCK_TIMER, "");
    sigcb(SIGINT, wrap (&halt), 0);

    main_worker = New worker(argc, argv);
    amain();
    // never reached
    exit(0);
}


