/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#ifndef _NODETABLE_H_
#define _NODETABLE_H_

#include "async.h"
#include "vec.h"
#include "sha1.h"
#include "skiplist.h"
#include "dqm_node.h"
#include "dqm_types.h"


typedef ip_port node_id;
typedef vec<node_id> node_list;
typedef enum {UP, DOWN} status;

class base_logger;

// In the future, we might want two classes: one that stores the nodes
// and another that maps keys to paths. For now, these two classes are the
// same because the original location table has some data structures that
// benefit from this conflation.
class nodetable {

  public:
    virtual ~nodetable() {}

    virtual void dump() = 0;

    // These three functions are the main interface to this object. Give
    // the first two functions a dqm_key, and they give the set of hosts
    // responsible for the key on either a logical put or a logical get. 
    // The third function is an optimization -- the 'put path' may be
    // derived from the 'get path'. The exact form of that derivation is
    // left to the sublcasses that inherit from this one, but the point
    // is that the object will calculate the 'put path' given the 'get
    // path'.

    virtual ptr<node_list> calc_path_get(ptr<dqm_key>) = 0;
    virtual ptr<node_list> calc_path_put(ptr<dqm_key>) = 0;
    virtual ptr<node_list> calc_path_put(ptr<node_list>) = 0;

    // This function is to be called *after* the node_id has already
    // been added with add_node. This function tells the node table who
    // the current node is. It is an error to call calc_path() without
    // having called this function.
    virtual void set_local_node(const node_id&) = 0;

    // These two functions, not surprisingly, add and remove nodes to
    // the table
    virtual void add_node (const node_id&, const str& alias_seed) = 0;
    virtual void remove_node(const node_id&) = 0;

    virtual status get_state(const node_id& id) = 0;
    virtual void set_state(const node_id& id, status stat) = 0;

    virtual void output_params(ptr<base_logger>) = 0;
 
  protected:
	// derived classes might find these structures and functions useful. 
	// no obligations, of course. 

	typedef u_int32_t node_alias;

	// a bunch of aliases that apply to a given node.
	typedef vec<node_alias> alias_vec;

	// goes inside a skiplist. 
	// these entries represent a node's existence in a particular ring.
	// the 'alias' is the node's alias in the ring. the 'id' is its
	// 'permanent' id.
	struct ringmember_lstelm {
	    node_alias alias;
	    node_id id;
	    sklist_entry<ringmember_lstelm> link;
	    bool operator< (ringmember_lstelm& other)
	    {
		return alias < other.alias;
	    }
	};

	// thanks for the skiplist, Emil!!
	typedef skiplist<ringmember_lstelm, 
	                 node_alias, 
	                 &ringmember_lstelm::alias,
		         &ringmember_lstelm::link> ring_lst;

	ringmember_lstelm* get_succ(node_alias alias, ring_lst* lst) {
	    return lst->closestsucc(alias);
	}

	// takes a node and returns a bunch of aliases.
        alias_vec get_node_aliases(const str& alias_seed, u_int32_t num) {

	    alias_vec aliases;
	    node_alias alias;
	    u_char hashed[sha1::hashsize];

	    for (u_int32_t i = 0; i < num; i++) {
		// construct a new string, e.g., "foo.3", "foo.4", etc.
		strbuf pre_hash("%s.%d", alias_seed.cstr(), i);
		str pre_hash_str(pre_hash);
		// . . . hash it . . .
		sha1_hash(hashed, pre_hash_str.cstr(), pre_hash_str.len());
		// . . . and use only the top bits . . .
		alias  = hash2alias(hashed);
		aliases.push_back(alias);
	    }
	    return aliases;
	}

	void dump_nodes_and_aliases(qhash<node_id,status>*     nodes, 
                                    qhash<node_id, alias_vec>* node_aliases) 
	{

	    qhash<node_id, status>::slot* s;
	    for (s = nodes->first(); s; s = nodes->next(s)) {

		in_addr addr;
		addr.s_addr = s->key.ip;
		warnx("node: %s:%u %s -- \n",
		    inet_ntoa(addr),
		    ntohs(s->key.port),
		    (s->value == UP) ? "UP" : "DOWN");

		// each node's aliases
		alias_vec* aliases = (*node_aliases)[s->key];
		for (u_int32_t j = 0; j < aliases->size(); j++)
		    warnx("%.6x ", (*aliases)[j]);

		warnx("\n");
	    }
	}

	virtual node_alias hash2alias(const u_char* hash) {
	    // we supply a default implementation in case the derived class
	    // doesn't need hash2alias or any similar functions but just
	    // wants to keep the interface.:w
	    return 0;
	}
};

#endif
