/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#include "arpc.h"
#include "async.h"
#include "sha1.h"
#include "dqm_portal.h"
#include "dqm_storer.h"
#include "portal.h"
#include "storer.h"
#include "nodetable.h"
#include "util.h"
#include "config.h"
#include "tt.h"
#include "log.h"
#include "stats.h"
#include "lru_cache.h"

//#define PRINTF_DEBUG

// timeout is 12 seconds (given by socket buffer size and disk seek
// latency). next request happens 0-2 seconds later, with
// the delay being randomly distributed.
//#define RPC_TMOUT      600
#define RPC_TMOUT      12    
#define MAX_JITTER_MS  1
#define STATS_INTERVAL 5

// global b/c portal, getter, and putter all need these variables
counters cntrs;
qhash<node_id, sockaddr_in> sockaddrs;

timecb_t* stats_cb = NULL;

void
portal::stats_printer_wrapper()
{
    stats_printer();
    stats_cb = delaycb(STATS_INTERVAL, 0, 
                         wrap(this, &portal::stats_printer_wrapper));
}

// We use sprintf b/c the %f code does not work for us with DM's
// strbuf objects. This might be a configuration problem, but, in any case,
// we work around it with sprintf. (Thanks, Max.)
#define LOG_TT(name, tt) \
    sprintf(entry, name ":%s:%d:%.2f", \
		    datebuf, \
		    cntrs.tt.counter(), \
		    cntrs.tt.num_per_sec());\
    str tt##str = entry; \
    logger->log(tt##str.cstr());

void
portal::stats_printer()
{
    char entry[100];
    char datebuf[60];
    sprintf(datebuf, "%.2f", get_now() / 1000000.);

    LOG_TT("TEST", tests);
    LOG_TT("SET", sets);
    LOG_TT("ATTGET", attempted_gets);
    LOG_TT("COMGET", completed_gets);
    LOG_TT("LOCGETATT", local_gets_attempted);
    LOG_TT("LOCGETCOM", local_gets_completed);
    LOG_TT("LOCGETTMO", local_gets_tmo);
    LOG_TT("LOCGETFND", local_gets_found);
    LOG_TT("ATTPUT", attempted_puts);
    LOG_TT("COMPUT", completed_puts);
    LOG_TT("LOCPUT", local_puts);
    LOG_TT("PUTTMOFIN", put_tmo);
    LOG_TT("GETTMOFIN", get_tmo);
    LOG_TT("LOCTESTFOUND", local_test_found);
    LOG_TT("LOCTESTNOTFOUND", local_test_not_found);
    LOG_TT("LOCSET", local_sets);
    LOG_TT("TESTYES", test_yes);
    LOG_TT("TESTNO", test_no);
    LOG_TT("RECVDGETS", received_gets);
    LOG_TT("RECVDPUTS", received_puts);

    logger->log("--");
}

portal::portal(const approved_list* storers,
	       u_int16_t portal_port,
               ptr<aclnt> client,
	       ptr<nodetable> nodetbl,
	       ptr<lru_cache> cache,
	       ptr<storer> loc_storer,
	       node_id local_id,
	       ptr<base_logger> lager,
	       bool* exit_token) :
    srv(init_udp_server(portal_port, dqm_portal_program_1)),
    nodetbl(nodetbl),
    cache(cache),
    local_storer(loc_storer),
    local_id(local_id),
    clnt(client),
    logger(lager),
    exit_token(exit_token)
{
    init_sockaddrs(storers);

    srv->setcb(wrap(this, &portal::dispatch));

    stats_cb = delaycb(STATS_INTERVAL, 0, 
                 wrap(this, &portal::stats_printer_wrapper));
    loc_storer->set_counters(&cntrs);
}

portal::~portal()
{
    stop();

    // print out stats one last time
    stats_printer();
    timecb_remove(stats_cb);

    *exit_token = true;
} 

// don't accept new requests any longer, but hold onto objects so we
// can complete the requests that were outstanding. to accomplish this,
// we tell the socket to stop receiving packets.
void 
portal::stop()
{  
    srv->xprt()->setrcb(NULL);
}

// populate the table of sockaddrs 
void
portal::init_sockaddrs(const approved_list* storers)
{
    approved_list::slot* s;

    for (s = storers->first(); s; s = storers->next(s)) {

	ip_port my_ip_port;
	my_ip_port.ip   = s->value.ip_addr;
	my_ip_port.port = htons(s->value.storer_port);

	sockaddr_in addr;
	memset(&addr, 0, sizeof(addr));
	addr.sin_family      = AF_INET;
	addr.sin_addr.s_addr = s->value.ip_addr;
	addr.sin_port        = htons(s->value.storer_port);

	sockaddrs.insert(my_ip_port, addr);
    }

}

void
portal::dispatch(svccb* rpc)
{
    if (!rpc) {
	srv->setcb(NULL);
	return;
    }

    switch (rpc->proc()) {

	case DQM_PORTAL_PROC_NULL: 
	    handle_NULL(rpc);
	    break;

	case DQM_PORTAL_PROC_TEST:
	    handle_TEST(rpc);
	    break;

	case DQM_PORTAL_PROC_SET:
	    handle_SET(rpc);
	    break;

	default:
	    fatal("Procedure %d not handled\n", rpc->proc());
    }

}

void
portal::handle_NULL(svccb* rpc)
{
    rpc->reply(NULL);
    warnx("got a NULL request\n");
}


void
portal::handle_TEST(svccb* rpc)
{
    cntrs.tests.inc();

    dqmportal_test_arg* arg = rpc->Xtmpl getarg<dqmportal_test_arg>(); 
    dqmportal_test_res res(DQM_OK);
    ptr<dqm_key> key = New refcounted<dqm_key>();
    memcpy(key->base(), arg->key.base(), arg->key.size());

#if 0
    warnx << *(sockaddr_in*)(rpc->getsa()) << "\n";
#endif

    // check locally first
    kv_rec* rec = cache->lookup(key, true);

    if (rec != NULL) {

	// found the (key,val) pair locally.
	memcpy(res.val->base(), rec->val->base(), rec->val->size());

	// NOTE: we trust ourselves, so we don't need to check the hash
	// locally. We know that we wouldn't have stored the (key,val) pair
	// if the invariants hadn't been satisfied.
	rpc->reply(&res);
	cntrs.local_test_found.inc();
	cntrs.test_yes.inc();
	return;
    }
    
    ptr<bool> called = New refcounted<bool>(false);
    timecb_t* tmo = delaycb(RPC_TMOUT, 0,
		      wrap(this, &portal::local_get_tmo, called, rpc, key));
    local_storer->local_get(key,
		   wrap(this, &portal::local_get_cb, called, tmo, rpc, key));

    cntrs.local_gets_attempted.inc();
}

// The purpose of these next two functions is to timeout if we're not
// called back in time. One or the other is executed, not both. (The
// first line of the second one could be executed along with the first
// one.)
void
portal::local_get_tmo(ptr<bool> called, svccb* rpc, ptr<dqm_key> key)
{
    *called = true;
    cntrs.local_gets_tmo.inc();
    handle_TEST_2(rpc, key, NULL);
}

void
portal::local_get_cb(ptr<bool> called, timecb_t* tmo, 
		     svccb* rpc, ptr<dqm_key> key, ptr<dqm_val> val)
{
    // we got called back too late; we've already moved on.
   if (*called) return;

    cntrs.local_gets_completed.inc();
    timecb_remove(tmo);
    handle_TEST_2(rpc, key, val);
}

void
portal::handle_TEST_2(svccb* rpc, ptr<dqm_key> key, ptr<dqm_val> loc_val)
{
    if (loc_val != NULL) {

	cntrs.local_gets_found.inc();

	// NOTE: we trust ourselves, so we don't need to check the hash
	// locally. We know that we wouldn't have stored the (key,val)
	// pair if the invariants hadn't been satisfied.
        ptr<dqmstorer_get_res> res = New refcounted<dqmstorer_get_res>(DQM_OK);
        memcpy(res->val->base(), loc_val->base(), loc_val->size());
	complete_TEST(rpc, key, true, res);

	return;

    } 

    // (key,val) not found locally. ask a subset of the other DQM nodes
    ptr<node_list> path = nodetbl->calc_path_get(key);
    if (!path->size()) {

	dqmportal_test_res res;
	// either we're the successor or no one else is up
	res.set_status(DQM_NOT_FOUND);
	rpc->reply(&res);
	cntrs.local_test_not_found.inc();
	cntrs.test_no.inc();
	return;
    }

    vNew getter (key, path, 
	         wrap(mkref(this), &portal::complete_TEST, rpc, key),
		 local_storer, local_id, clnt, logger);

}


void
portal::complete_TEST(svccb* rpc, ptr<dqm_key> key, 
                      bool found, ptr<dqmstorer_get_res> val) 
{
    dqmportal_test_res res;
    if (found) {

	// Cache locally
	ptr<dqm_val> v = New refcounted<dqm_val> ();
	memcpy(v->base(), val->val->base(), val->val->size());

	cache->add(key, v);

	res.set_status(DQM_OK);

	memcpy(res.val->base(), val->val->base(), val->val->size());
	rpc->reply(&res);

	cntrs.test_yes.inc();

    } else {

	res.set_status(DQM_NOT_FOUND);
	rpc->reply(&res);
	cntrs.test_no.inc();

    }
}

void
portal::handle_SET(svccb* rpc)
{
    cntrs.sets.inc();

    dqmportal_set_arg* arg = rpc->Xtmpl getarg<dqmportal_set_arg>();
    ptr<dqm_key> key = New refcounted<dqm_key> ();
    ptr<dqm_val> val = New refcounted<dqm_val> ();

    memcpy(val->base(), arg->val.base(), arg->val.size());
    sha1_hash(key->base(), val->base(), val->size());

    if (memcmp(key->base(), arg->key.base(), arg->key.size())) {

	rpc->replyref(dqm_stat(DQM_BAD_PARAM));
	return;
    }

    // cache the stamp locally
    cache->add(key, val);

    // always do a local put.
    local_storer->local_put(key, val);
    cntrs.local_puts.inc();

    ptr<node_list> path = nodetbl->calc_path_put(key);

    if (!path->size()) {

#ifdef PRINTF_DEBUG
	warnx << "in SET, path->size() == 0\n";
#endif
	// we're the successor or no one else was up. get out of here.
	cntrs.local_sets.inc();
	rpc->replyref(dqm_stat(DQM_OK));
	return;
    }

#ifdef PRINTF_DEBUG
    warnx << "about to invoke putter\n";
#endif

    vNew putter (key, val, path,	    
		 wrap(mkref(this), &portal::complete_SET, rpc),
		 local_storer, local_id,
		 clnt, logger);
}

void
portal::complete_SET(svccb* rpc, u_int32_t rc)
{
    // if (rc != 0) {
    //    strbuf sb;
    //    sb << "stamp put had some errors: " << rc;
    //    str s(sb);
    //    logger->log(s.cstr());
    rpc->replyref(dqm_stat(rc));
}

getter::getter(ptr<dqm_key> key,  
	       ptr<node_list> path, getter_cb cb,
	       ptr<storer> local_storer,
	       node_id  local_id,
	       ptr<aclnt> clnt, ptr<base_logger> logger) :
    key(key),
    cb(cb),
    path(path),
    pos(0),
    local_storer(local_storer),
    local_id(local_id),
    clnt(clnt),
    logger(logger)
{
    make_get_req();
}

getter::~getter() 
{
}

// to avoid synchronization, make the request after a randomly-distributed
// bit of delay
void
getter::make_get_req()
{
    // jitter if we're trying any but the first one on the path
    if (pos > 0) {
	int jitter = get_rand_int(0, MAX_JITTER_MS);
	int sec = jitter / 1000;
	int nsec = (jitter % 1000)*1000000;	 // convert ms to ns
	delaycb(sec, nsec, wrap(this, &getter::make_get_req_worker));
	//warnx ("delay: sec: %d. nsec: %d\n", sec, nsec);
    } else {
	make_get_req_worker();
    }
}

void
getter::make_get_req_worker()
{
    node_id next = (*path)[pos];

    if (next == local_id) {

	// skip, because we already did a local get, above.
	iterate();

    } else {

	dqmstorer_get_arg arg;
	ptr<dqmstorer_get_res> res = New refcounted<dqmstorer_get_res>();

	memcpy(arg.key.base(), key->base(), key->size());

	sockaddr_in* addr = sockaddrs[next];

	clnt->timedcall((time_t)RPC_TMOUT,	  // timeout in seconds
			DQM_STORER_PROC_GET,  // procno
			&arg,                 // inmem
			res,		  // outmem
			wrap(this, &getter::get_cb, res),    // callback
			NULL,		  // auth
			NULL,		  // inproc
			NULL,		  // outproc
			0,			  // progno
			0,			  // versno
			(sockaddr*)addr);

#ifdef PRINTF_DEBUG
	warnx << "making a remote get call\n";
#endif
	cntrs.attempted_gets.inc();
    }
}

void
getter::get_cb(ptr<dqmstorer_get_res> res, clnt_stat err)
{
    if (process_rpc_reply(res, err))
	iterate();
}


// returns true if we should keep iterating, false otherwise
bool
getter::process_rpc_reply(ptr<dqmstorer_get_res> res, clnt_stat err)
{
    char hashbuf[sha1::hashsize];

    switch (err) {

    case RPC_SUCCESS:

	cntrs.completed_gets.inc();

	if (res->status == DQM_OK)  {

	    sha1_hash(hashbuf, res->val->base(), res->val->size());
	   
	    if (!memcmp(hashbuf, key->base(), sha1::hashsize)) {

#ifdef PRINTF_DEBUG
		warnx << "remote RPC worked. replying...\n";
#endif
		(*cb)(true, res);
		delete this;
		return false;

	    } else {
		strbuf sb;
		sb << "remote storer "
		   << (*path)[pos]
		   << " gave bad value " 
		   << hexdump(res->val->base(), res->val->size())
		   << " on key "
		   << hexdump(key->base(), key->size());
		str s(sb);
		logger->log(s.cstr());
		// keep going
	    }
	}  
	// the key wasn't found; keep going
	break;

    case RPC_TIMEDOUT:

	cntrs.get_tmo.inc();
	break;

    default:
	strbuf sb;
	sb << "GET RPC call to node " 
	   << (*path)[pos] 
	   << " failed on key "
	   << hexdump(key->base(), key->size())
	   << " with error: "
	   << clnt_sperrno(err);
	str s(sb);
	logger->log(s);
    }

#ifdef PRINTF_DEBUG
    warnx << "continuing on get\n";
#endif

    return true;
}

void
getter::iterate()
{
    if (pos < path->size() - 1) {

	pos++;
	make_get_req();

    } else  {

	(*cb)(false, NULL);
	delete this;
	return;
    }
}

putter::putter(ptr<dqm_key> key, 
	       ptr<dqm_val> val, 
	       ptr<node_list> path, 
	       putter_cb cb,
	       ptr<storer> local_storer,
	       node_id     local_id,
	       ptr<aclnt> clnt,
	       ptr<base_logger> logger) :
    key(key),
    val(val),
    path(path),
    pos(0),
    cb(cb),
    errs(0),
    local_storer(local_storer),
    local_id(local_id),
    clnt(clnt),
    logger(logger)
{
    make_put_req();
}

putter::~putter() 
{
}

void
putter::make_put_req()
{
    // jitter if we're trying any but the first one on the path
    if (pos > 0) {
	int jitter = get_rand_int(0, MAX_JITTER_MS);
	int sec = jitter / 1000;
	int nsec = (jitter % 1000)*1000000;	 // convert ms to ns
	delaycb(sec, nsec, wrap(this, &putter::make_put_req_worker));
	//warnx ("delay: sec: %d. nsec: %d\n", sec, nsec);
    } else {
	make_put_req_worker();
    }
}

void
putter::make_put_req_worker()
{
    node_id next = (*path)[pos];

    if (next == local_id) {

	// skip, because we already did a local put, above.
	iterate();

    } else {

	dqmstorer_put_arg arg;
	dqm_stat* res = New dqm_stat;

	memcpy(arg.key.base(), key->base(), key->size());
	memcpy(arg.val.base(), val->base(), val->size());
	arg.ttl = 0;

	sockaddr_in* addr = sockaddrs[next];

	clnt->timedcall((time_t)RPC_TMOUT,	  // timeout in seconds
			DQM_STORER_PROC_PUT,  // procno
			&arg,                 // inmem
			res,		  // outmem
			wrap(this, &putter::put_cb, res),    // callback
			NULL,		  // auth
			NULL,		  // inproc
			NULL,		  // outproc
			0,			  // progno
			0,			  // versno
			(sockaddr*)addr);

	cntrs.attempted_puts.inc();
#ifdef PRINTF_DEBUG
	warnx << "making remote put\n";
#endif
    }
}

void
putter::put_cb(dqm_stat* res, clnt_stat err)
{
    dqm_stat res_copy = *res;
    delete res;

    process_rpc_reply(res_copy, err);
    iterate();
}

void
putter::process_rpc_reply(dqm_stat res_copy, clnt_stat err)
{
    switch (err) {

    case RPC_SUCCESS:

	cntrs.completed_puts.inc();

	if (res_copy != DQM_OK) 
	    errs = errs | res_copy;
	break;

    case RPC_TIMEDOUT:

	cntrs.put_tmo.inc();
	break;

    default:
	strbuf sb;
	sb << "PUT RPC to node " 
	   << (*path)[pos] 
	   << " on key "
	   << hexdump(key->base(), key->size())
	   << " didn't complete: "
	   << clnt_sperrno(err);
	str s(sb);
	logger->log(s.cstr());
    }
}

void
putter::iterate()
{
    if (pos < path->size() - 1) {

#ifdef PRINTF_DEBUG
	warnx << "iterating to next put\n";
#endif
	pos++;
	make_put_req();

    } else  {

#ifdef PRINTF_DEBUG
	warnx << "calling back client\n";
#endif
	(*cb)(errs);
	delete this;

    }
}

