/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <db.h>
#include "arpc.h"
#include "async.h"
#include "sha1.h"
#include "dqm_storer.h"
#include "storer.h"
#include "util.h"
#include "config.h"
#include "log.h"
#include "stats.h"
#include "ondisk_storage.h"
#include "lru_cache.h"

storer::storer(u_int16_t storer_port, 
	       ptr<nodetable> nodetbl,
	       ptr<ondisk_storage> storage,
	       ptr<lru_cache> cache,
	       ptr<base_logger> logger) :
   srv(init_udp_server(storer_port, dqm_storer_program_1, axprt::defps*20)), 
   nodetbl(nodetbl),
   cache(cache),
   storage(storage),
   logger(logger),
   cntrs(NULL)
{
    srv->setcb(wrap(this, &storer::dispatch));
}

storer::~storer()
{
    stop();
    storage->stop();
}

// don't accept new requests any longer, but hold onto objects so we
// can complete the requests that were outstanding. to accomplish this,
// we tell the socket to stop receiving packets.
void
storer::stop()
{
    srv->xprt()->setrcb(NULL);
}

void
storer::dispatch(svccb* rpc)
{
    if (!rpc) {
	srv->setcb(NULL);
	return;
    }

    switch (rpc->proc()) {

	case DQM_STORER_PROC_NULL: 
	    handle_NULL(rpc);
	    break;

	case DQM_STORER_PROC_GET:
	    handle_GET(rpc);
	    break;

	case DQM_STORER_PROC_PUT:
	    handle_PUT(rpc);
	    break;

	default:
	    fatal("Procedure %d not handled\n", rpc->proc());
    }

}

void
storer::handle_NULL(svccb* rpc)
{
    rpc->reply(NULL);
    logger->log("got a NULL request\n");
}

void
storer::handle_GET(svccb* rpc)
{
    dqmstorer_get_arg* arg = rpc->Xtmpl getarg<dqmstorer_get_arg>(); 

    cntrs->received_gets.inc();

    ptr<dqm_key> key = New refcounted<dqm_key> ();
    memcpy(key->base(), arg->key.base(), arg->key.size());
    
    get_worker(key, wrap(mkref(this), &storer::complete_GET, rpc));
}

// XXX: should only accept if we're responsible for the key
void
storer::handle_PUT(svccb* rpc)
{
    dqmstorer_put_arg* arg = rpc->Xtmpl getarg<dqmstorer_put_arg>();

    if (arg->key.size() != HASH_SIZE ||
        arg->val.size() != HASH_SIZE) {
	strbuf sb;
	sb << "got a put request from "
          << *(sockaddr_in*)rpc->getsa() 
	  << " with bad key or value size\n";
	str s(sb);
	logger->log(s.cstr());
	return;
    }

    cntrs->received_puts.inc();

    ptr<dqm_val> val = New refcounted<dqm_val> ();
    memcpy(val->base(), arg->val.base(), arg->val.size());

    ptr<dqm_key> key = New refcounted<dqm_key> ();
    sha1_hash(key->base(), val->base(), val->size());
    if (memcmp(key->base(), arg->key.base(), arg->key.size())) {
	rpc->replyref(dqm_stat(DQM_BAD_PARAM));
	return;
    }
 
    put_worker(key, val);
    rpc->replyref(dqm_stat(DQM_OK));
}

void
storer::local_get(ptr<dqm_key> key, getcb cb)
{
    get_worker(key, cb);
}

void
storer::get_worker(ptr<dqm_key> key, getcb cb)
{
    if (kv_rec* rec = cache->lookup(key, true)) {
	(*cb)(rec->val);
    } else {
	storage->get(key, wrap(mkref(this), &storer::get_worker_2, key, cb));
    }
}

void
storer::get_worker_2(ptr<dqm_key> key, getcb cb, ptr<dqm_val> val)
{
    if (val != NULL)  
	cache->add(key, val);

    (*cb)(val);
}

void
storer::complete_GET(svccb* rpc, ptr<dqm_val> val)
{
    dqmstorer_get_res res(DQM_OK);
    
    if (val == NULL)
	res.set_status(DQM_NOT_FOUND);
    else
	memcpy(res.val->base(), val->base(), val->size());

    rpc->reply(&res);
}

void
storer::local_put(ptr<dqm_key> key, ptr<dqm_val> val)
{
    put_worker(key, val);
}

void
storer::put_worker(ptr<dqm_key> key, ptr<dqm_val> val)
{
    cache->add(key, val);
    storage->put(key, val);  
}

void
storer::set_counters(counters* cs)
{
    cntrs = cs;
}
