/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#ifndef _TT_H_
#define _TT_H_

#include <sys/time.h>

/* us is the number of microseconds since the epoch. */
inline void us2tv(u_int64_t us, struct timeval* tv)
{
    tv->tv_sec  = us / 1000000;
    tv->tv_usec = us - tv->tv_sec*1000000LL;
}

inline void us2tv(u_int32_t us, struct timeval* tv)
{
    tv->tv_sec  = us / 1000000;
    tv->tv_usec = us % 1000000;
}

inline u_int64_t get_now()
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv.tv_sec*1000000LL + tv.tv_usec;
}

// keep track of the avg. amount of time to do a particular task.
// callers should call begin() and end() before and after the task,
// respectively. avg() gives the average amount of time the task has been
// taking.
class time_it {

    public:
	time_it () : total(0), cntr(0), curr_min(0) {}
	~time_it() {}
	
	void begin()   {
	    time_at_begin = get_now();
	}
	void end() {
	    u_int64_t delta = get_now() - time_at_begin;
	    total += delta;
	    if (cntr == 0 || delta < curr_min)
		curr_min = delta;
	    ++cntr;
	}
	double avg_us() {
	    if (cntr == 0) return -1.;
	    return (double)total / cntr;
	}

	double min_us() {
	    if (cntr == 0) return -1.;
	    return curr_min;
	}

    private:
	u_int64_t total;	
	u_int32_t cntr;
	u_int64_t time_at_begin;
	u_int64_t curr_min;
};

// keep track of the rate at which a counter is being incremented
class tt {

    public:
	tt() : last_time(get_now()), last_counter(0), curr_counter(0) {}
	~tt() {}

	u_int32_t counter() { return curr_counter; }
	void inc() { ++curr_counter; }

	double num_per_sec() {

	    u_int64_t curr_time = get_now();

	    // calculate the gaps
	    u_int32_t num_in_interval = curr_counter - last_counter;
	    u_int64_t delta           = curr_time - last_time;
	   
	    // update our internal state
	    last_time    = curr_time;
	    last_counter = curr_counter;

	    // and do the calculation
	    double rate = (num_in_interval * 1000000.) / delta;

	    return rate;
	}

    private:

	u_int64_t last_time;  // in usec
	u_int32_t last_counter;
	u_int32_t curr_counter;

};

#endif
