/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include "vec.h"
#include "rxx.h"
#include "arpc.h"
#include "util.h"

#define PATH_DELIM  '/'

// try to find the largest socket buffer size that works and then "fool"
// DM's alloc function by passing in a socket and a requested size that
// match.
size_t
setrcvsockbuf(int fd, size_t attempted_sz, size_t inc)
{
    size_t actual_sz = 0;
    size_t old_actual;
    socklen_t sn = sizeof (actual_sz);
    if (getsockopt (fd, SOL_SOCKET, SO_RCVBUF, &actual_sz, &sn) < 0)
	return attempted_sz;

    old_actual = actual_sz;
    
    // quit when either the actual_sz is greater than what we attempted
    // or when the kernel actually adjusted the size (even if it didn't
    // complain in the setsockopt). Note that if the attempted_sz gets
    // smaller and smaller, the first condition in the while loop will
    // let us exit without *shrinking* the socket buffer.
    while (attempted_sz > actual_sz && old_actual == actual_sz) {

	if (setsockopt (fd, SOL_SOCKET, SO_RCVBUF, 
	    &attempted_sz, sizeof (attempted_sz)) < 0) 
	       attempted_sz -= inc;

	old_actual = actual_sz;
	getsockopt(fd, SOL_SOCKET, SO_RCVBUF, &actual_sz, &sn);
    }
    return actual_sz;
}

/* supply the port in host order. */
ptr<asrv>
init_udp_server(uint16_t port, const rpc_program& program, 
	        size_t attempted_sz)
{
    int dgram_fd = inetsocket (SOCK_DGRAM, port);
    if (dgram_fd < 0) 
	fatal("Failed to allocate server datagram socket bound "
	      "to port %d\n", port);

    size_t actual_sz = setrcvsockbuf(dgram_fd, attempted_sz, axprt::defps);
    ptr<axprt_dgram> dx = axprt_dgram::alloc (dgram_fd, sizeof(sockaddr), 
					      actual_sz);
    if (!dx) fatal << "Failed to allocate dgram xprt\n";

    ptr<asrv> srv = asrv::alloc (dx, program, NULL);
    if (!srv) fatal << "Failed to allocate server object\n";

    return srv;
}

/* 
 * supply the port in host order. this is the port *from* which
 * RPC messages will be originated.
 */
ptr<aclnt>
init_udp_client(uint16_t port, const rpc_program& program)
{
    ptr<aclnt> clnt;

    // get a larger socket buffer size because we don't ever want to
    // lose replies to RPCs; doing so is equivalent to wasting work.
    size_t attempted_sz = axprt::defps*20;

    int dgram_fd = inetsocket (SOCK_DGRAM, port);
    if (dgram_fd < 0) 
	fatal("Failed to allocate client datagram socket bound "
	      "to port %d\n", port);
  
    size_t actual_sz = setrcvsockbuf(dgram_fd, attempted_sz, axprt::defps);
    ptr<axprt_dgram> dx = axprt_dgram::alloc (dgram_fd, sizeof(sockaddr),
                                              actual_sz);
    if (!dx) fatal << "Failed to allocate dgram xprt\n";

    // The fourth argument to the alloc function asks the aclnt to use
    // a message buffer that doesn't get queued up and retried. DM's
    // default is to retry 5 times, with exponential backoff. With this
    // fourth argument, we are asking the code to try once and, if the
    // transmission is lost, that's it -- no more retries.
    clnt = aclnt::alloc (dx, program, NULL, callbase_alloc<rpccb_msgbuf_xmit>);

    if (!clnt) fatal << "Failed to allocate client object\n";

    return clnt;
}


// got this code from Max Krohn. thanks, Max.
static rxx slash ("/+");

bool
makedirs(const char* dir)
{
  vec<str> subdirs;
  int rc;
  if ((rc = split (&subdirs, slash, dir)) <= 0) 
    return false;

  strbuf b;

  if (dir[0] == '/')
    b << "/";

  for (u_int i = 0; i < subdirs.size (); i++) {
    if (subdirs[i].len ()) {
      b << subdirs[i];
      if (mkdir (str (b), 0777) && errno != EEXIST) 
         return false;
      b << "/";
    }
  }
  return true;
}

// returns true if error happens
bool
convertdouble(const char* cp, double* resp)
{
    char* dummy;
    *resp = strtod(cp, &dummy);
    return (dummy == cp);
}


