/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
//
// $Id: calqueue.h,v 1.6 2006/05/19 13:17:03 mwalfish Exp $
//
#ifndef _CALQUEUE_H_
#define _CALQUEUE_H_

#include <string.h>
#include <stdlib.h>
#include <assert.h>

// A calender queue that stores objects of type T. T needs to have:
//   -- a when() function defined 
//   -- a constructor that can be called with 0, like this: T(0)
//      (integers or pointers already have this.)

template <typename T>
class CalQueue {

    template<typename S> struct _wrapper {
	S t;
	struct _wrapper* next;
    };
    typedef struct _wrapper<T> wrapper;

    typedef unsigned long long uint64;

    public:
	CalQueue(int year_usec);
	~CalQueue();

	void insert(T); 

	// the next two functions extract, or inspect, the item that occurs
	// soonest after curr. if there are no pending events, the next
	// function returns T(0).
	T extract_min(uint64 curr);
	T inspect_min(uint64 curr);

	void set_yearsize(int year_usec)  { m_year_usec = year_usec; }
	int  year()       { return m_year_usec;   }
	int  num_slots()  { return m_nslots;      }
	int  size()       { return m_size;        }

	// for testing and debugging; not really public.
	T* dump() { return m_array; }

    private:

	// code common to extract_min and inspect_min
	int get_min_slot(uint64 curr);
	// helper function
	void insert(wrapper** arr, wrapper* item, 
		    int num_slots, int day_len_usec);

	void recalc_slots(int num_slots);

	int       m_size;	// number of items in the queue
	int       m_nslots;     // analogy is to days on the calendar
	int       m_day_usec;   // length of a day in microseconds
	int       m_year_usec;  // analogy is to year on the calendar
	wrapper** m_array;	// the slots
};

template <typename T> 
CalQueue<T>::CalQueue(int year_usec) :
    m_size(0), 
    m_nslots(1), 
    m_day_usec(year_usec),
    m_year_usec(year_usec),
    m_array( (wrapper**)calloc(m_nslots, sizeof(wrapper*)) )
{
}

template <typename T>
CalQueue<T>::~CalQueue()
{
    free(m_array);
}

template <typename T>
void CalQueue<T>::recalc_slots(int num_slots)
{
    int day_len_usec = m_year_usec / num_slots;
    printf("recalculating slots. num: %d. width: %d\n", 
           num_slots, day_len_usec);

    wrapper** new_slots = (wrapper**)calloc(num_slots, sizeof(wrapper*));
    wrapper* curr;

    // pull all of the current items off of the queue and reinsert them
    // in the new queue.
    for (int i = 0; i < m_nslots; i++) {

	curr = m_array[i];

	while (curr) {
	    m_array[i] = curr->next;
	    curr->next = NULL;
	    insert(new_slots, curr, num_slots, day_len_usec);
	    curr = m_array[i];
	}
    }

    free(m_array);
    m_array     = new_slots;
    m_nslots    = num_slots;
    m_day_usec  = day_len_usec;
}

template <typename T>
inline void CalQueue<T>::insert(T to_insert)
{
    if (m_size >= 2*m_nslots) 
	recalc_slots(2*m_nslots);

    wrapper* wr = (wrapper*)malloc(sizeof(wrapper));
    wr->t    = to_insert;
    wr->next = NULL;

    insert(m_array, wr, m_nslots, m_day_usec);
    ++m_size;
}

template <typename T> 
inline T CalQueue<T>::extract_min(uint64 from)
{
    if (m_size < 1) 
	return T(0);

    if (m_size <= m_nslots / 2)
	recalc_slots(m_nslots / 2);

    int slot = get_min_slot(from);

    wrapper* curr = m_array[slot];
    m_array[slot] = curr->next;
    T ret = curr->t;
    free(curr);
    --m_size;

    return ret;   

}

template <typename T> 
inline T CalQueue<T>::inspect_min(uint64 from)
{
    if (m_size < 1)
	return T(0);

    int slot = get_min_slot(from);
    return m_array[slot]->t;
}

template <typename T>
inline int CalQueue<T>::get_min_slot(uint64 curr)
{
    uint64 slot_end;
    uint64 absolute_slot = curr / m_day_usec;
    uint64 slot_beg      = absolute_slot * m_day_usec;
    int array_slot       = absolute_slot % m_nslots; 

    while (1) {

	slot_end = slot_beg + m_day_usec;

	// If the current index has an entry in the current 'time
	// slot' (a period of length m_day_usec), return that entry. 
	// If not, keep cycling through the events.
	if (m_array[array_slot] && m_array[array_slot]->t.when() < slot_end) 
	    break;	   

	slot_beg = slot_end;
	array_slot = (slot_beg / m_day_usec) % m_nslots;
    } 

    return array_slot;
}

// helper function. should be friend.
template <typename T>
inline void CalQueue<T>::insert(wrapper** array, wrapper* item, 
	                 int num_slots, int day_len_usec)
{
    uint64 when = item->t.when();
    int array_slot = (when / day_len_usec) % num_slots;
    wrapper** curr;

    // insert new element in order in the appropriate list
    if (array[array_slot] == NULL) {
	array[array_slot] = item; 
	return;
    }

    for (curr = &array[array_slot]; *curr; curr = &(*curr)->next) {
	if (when < (*curr)->t.when()) {
	    item->next = *curr;
	    *curr = item;
	    break;
	}
    }

    if (*curr == NULL) 
	*curr = item;
}


#endif
