/*
 * Copyright (C) 2005-2006 Michael Walfish
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
/*
 * $Id: send.h,v 1.27 2006/05/19 13:17:03 mwalfish Exp $
 *
 */
#ifndef _SEND_H_
#define _SEND_H_

#include <map>
#include <ext/hash_map>
#include "u.h"
#include "heap.h"

#define OUR_WELL_KNOWN_PORT   2100
#define METASTAMP_MAX_SIZE    1024
#define DUMP_INTERVAL_SEC       1
#define RPC_TIMEOUT      600

/* sample one out of every 100 requests */
#define LATENCY_SAMPLE    100

/* us is the number of microseconds since the epoch. */
inline void us2tv(uint64 us, struct timeval* tv)
{
    tv->tv_sec  = us / 1000000;
    tv->tv_usec = us - tv->tv_sec*1000000LL;
}

inline void us2tv(uint32 us, struct timeval* tv)
{
    tv->tv_sec  = us / 1000000;
    tv->tv_usec = us % 1000000;
}

inline uint64 get_now()
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv.tv_sec*1000000LL + tv.tv_usec;
}

/* structures for working with the event queue  */
typedef enum { TEST = 1, DUMP = 2, TEST_TMO = 3, PRELOAD_DONE = 4, DONE_RUNNING = 5, RATE_UPDATE = 6 } proc_type;

struct event {

    event()                     : wh(0), type(TEST), xid(0) {}
    event(uint64 w)             : wh(w), type(TEST), xid(0) {}
    event(uint64 w, proc_type t): wh(w), type(t),    xid(0) {}

    uint64 when()      { return wh; }

    uint64 wh;
    proc_type type;
    int xid;	// only used for outstanding tests
};

typedef struct _outstanding_test {
   bool    is_spam;
   uint32  n_spam;
    /* 
     * To save memory, we only store the v part of <k,v> on our 
     * to-do list. The SET RPC recalculates k. If the CPU turns out
     * to be too busy and we have extra memory, we can store k too.
     */
   uint8   val[KEY_VAL_LEN];
   portal* poor;
   uint64  begin;
   uint64  when;

} outstanding_test;

typedef struct _stamp_cntr {

    uint32 att;
    uint32 found;
    uint32 notfound;
    uint32 tmo;
} stamp_cntr;

typedef __gnu_cxx::hash_map<int, outstanding_test*> xidmap_t;
typedef xidmap_t::iterator xid_it;

typedef std::multimap<uint64, event>  events;
typedef events::iterator              ev_it;
typedef std::pair<uint64, event>      ev_pair; 


/* 
 * Constants for working with the portal's RPC server. Note that 
 * we are not using the standard XDR format but rather one with 16
 * fewer bytes.
 */
#define RPC_CALL         0
#define RPC_REPLY        1
/* reply_stat */
#define RPC_MSG_ACCEPTED 0
#define RPC_MSG_DENIED   1
/* accept_stat */
#define RPC_SUCCESS      0
/* dqm error codes */
#define DQM_OK                  0
#define DQM_NOT_FOUND           1
#define DQM_BAD_METASTAMP       2
#define DQM_METASTAMP_REUSED    4
#define DQM_SEEN_KEY            8
#define DQM_NO_MORE_STORAGE    16
#define DQM_BAD_PARAM          32

#define RPC_VERSION      2
#define PORTAL_PROG      23094122
#define PORTAL_PROG_VER  1
#define PORTAL_TEST_PROC 2
#define PORTAL_SET_PROC  3

#define TEST_RPC_SIZE  (sizeof(portal_req_hdr) + sizeof(test_req_body))

typedef struct _portal_req_hdr {
    int xid;
    int type;
    int rpc_version;
    int prog;
    int prog_vers;
    int proc;
    int auth_dummy[4];
} portal_req_hdr;

typedef struct _test_req_body {
    uint8 key[KEY_VAL_LEN];
} test_req_body;

typedef struct _set_req_body {
    uint8 key[KEY_VAL_LEN];
    uint8 val[KEY_VAL_LEN];
    int ms_len;
    char  metastamp[METASTAMP_MAX_SIZE];  /* rough guess */
} set_req_body;

typedef struct _portal_req {
    portal_req_hdr hdr;
    union {
	test_req_body test;
	set_req_body  set;
    } body;
} portal_req;

typedef struct _portal_resp_hdr {
    int xid;
    int type;
    int reply_stat;
    int auth_dummy[2];
    int accept_stat;
} portal_resp_hdr;

typedef struct _portal_resp_body {
    int res;
    uint8 val[KEY_VAL_LEN];
} portal_resp_body;

typedef struct _portal_resp {
    portal_resp_hdr hdr;
    portal_resp_body body;
} portal_resp;

#define RESP_HDR_SIZE           (sizeof(portal_resp_hdr))
#define TEST_RESP_FOUND_SIZE    (sizeof(portal_resp))
#define TEST_RESP_NOTFOUND_SIZE (sizeof(portal_resp) - KEY_VAL_LEN)
#define SET_RESP_SIZE           (sizeof(portal_resp) - KEY_VAL_LEN)

struct config_params {
    portals* ps;
    double  pr_spam;
    uint32  universe_sz;
    uint32  msgc_begin;
    uint32  msgc_max;
    int     repeat;
    int     rate;
    int     preload_rate;
    bool    done_delay;
    bool    tests_only;
    bool    sets_only;
    char*   latency_file;  
    char*   rate_file;
    uint32  rate_file_o;
    uint64  rate_file_t;
};

void sendloop(config_params*);

void stop_running();

struct counters {
    uint32 set_req;
    uint32 set_resp;
    uint32 test_req;
    uint32 test_found;
    uint32 test_notfound;
    uint32 test_tmo;
    uint32 test_resp_after_tmo;
    uint32 hash_check_failed;
};


#endif
